from unittest.mock import patch
from uuid import UUID

from django.db import connections
from django.db.backends.base.base import _thread

from arkindex.documents.models import ElementPath
from arkindex.documents.tasks import move_element
from arkindex.project.tests import FixtureTestCase

PATHS_IDS = [
    UUID('00000000-0000-0000-0000-000000000000'),
    UUID('11111111-1111-1111-1111-111111111111'),
    UUID('22222222-2222-2222-2222-222222222222'),
    UUID('33333333-3333-3333-3333-333333333333'),
    UUID('44444444-4444-4444-4444-444444444444'),
    UUID('55555555-5555-5555-5555-555555555555'),
    UUID('66666666-6666-6666-6666-666666666666'),
    UUID('77777777-7777-7777-7777-777777777777'),
    UUID('88888888-8888-8888-8888-888888888888'),
]


class TestMoveElement(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page_type = cls.corpus.types.get(slug='page')
        cls.destination = cls.corpus.elements.get(name='Volume 2')
        cls.parent = cls.corpus.elements.get(name='Volume 1')
        cls.source_with_children = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.source_without_child = cls.corpus.elements.get(name='Volume 1, page 2r')
        ElementPath.objects.filter(path__contains=[cls.source_without_child.id]).delete()

    @patch.object(ElementPath._meta.get_field('id'), 'get_default')
    def test_run_on_source_without_child(self, default_field_mock):
        default_field_mock.return_value = PATHS_IDS[0]

        # No child on this page
        self.assertEqual(ElementPath.objects.filter(path__contains=[self.source_without_child.id]).count(), 0)

        source_paths = ElementPath.objects.filter(element_id=self.source_without_child.id)
        self.assertEqual(len(source_paths), 1)
        self.assertEqual(list(source_paths.values('path')), [{'path': [self.parent.id]}])

        with self.assertExactQueries('element_move_without_child.sql', params={
            'source_id': str(self.source_without_child.id),
            'parent_id': str(self.parent.id),
            'destination_id': str(self.destination.id),
            'page_type_id': str(self.page_type.id),
            'path_id': str(PATHS_IDS[0]),
            'savepoints': [f"s{_thread.get_ident()}_x{connections['default'].savepoint_state + 1}", f"s{_thread.get_ident()}_x{connections['default'].savepoint_state + 2}"]
        }):
            move_element(self.source_without_child, self.destination)

        self.assertEqual(len(source_paths), 1)
        self.assertEqual(list(source_paths.values('path')), [{'path': [self.destination.id]}])

    @patch.object(ElementPath._meta.get_field('id'), 'get_default')
    def test_run_on_source_with_children(self, default_field_mock):
        default_field_mock.side_effect = PATHS_IDS

        # 4 children on this page
        children_paths = ElementPath.objects.filter(path__contains=[self.source_with_children.id])
        self.assertEqual(children_paths.count(), 4)
        self.assertEqual(list(children_paths.values('path')), [
            {'path': [self.parent.id, self.source_with_children.id]},
            {'path': [self.parent.id, self.source_with_children.id]},
            {'path': [self.parent.id, self.source_with_children.id]},
            {'path': [self.parent.id, self.source_with_children.id]}
        ])

        source_paths = ElementPath.objects.filter(element_id=self.source_with_children.id)
        self.assertEqual(len(source_paths), 1)
        self.assertEqual(list(source_paths.values('path')), [{'path': [self.parent.id]}])

        with self.assertExactQueries('element_move_with_children.sql', params={
            'source_id': str(self.source_with_children.id),
            'parent_id': str(self.parent.id),
            'destination_id': str(self.destination.id),
            'children_ids': [str(id) for id in children_paths.values_list('element_id', flat=True)],
            'page_type_id': str(self.page_type.id),
            'paths_ids': [str(id) for id in PATHS_IDS],
            'savepoints': [f"s{_thread.get_ident()}_x{connections['default'].savepoint_state + 1}", f"s{_thread.get_ident()}_x{connections['default'].savepoint_state + 2}"]
        }):
            move_element(self.source_with_children, self.destination)

        self.assertEqual(len(source_paths), 1)
        self.assertEqual(list(source_paths.values('path')), [{'path': [self.destination.id]}])

        # Assert children were also moved
        self.assertEqual(list(children_paths.values('path')), [
            {'path': [self.destination.id, self.source_with_children.id]},
            {'path': [self.destination.id, self.source_with_children.id]},
            {'path': [self.destination.id, self.source_with_children.id]},
            {'path': [self.destination.id, self.source_with_children.id]}
        ])
