from django.db.models import Q
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiExample, OpenApiParameter, extend_schema, extend_schema_view
from rest_framework import status
from rest_framework.generics import CreateAPIView, ListAPIView, RetrieveUpdateAPIView
from rest_framework.response import Response
from rest_framework.serializers import ValidationError

from arkindex.documents.models import Corpus, Element
from arkindex.documents.serializers.elements import ElementSlimSerializer
from arkindex.images.models import Image
from arkindex.images.serializers import (
    CreateImageErrorResponseSerializer,
    IIIFImageSerializer,
    ImageInformationSerializer,
    ImageSerializer,
    ImageUploadSerializer,
)
from arkindex.project.aws import S3FileStatus
from arkindex.project.permissions import IsVerified
from arkindex.users.models import Scope


@extend_schema_view(
    post=extend_schema(
        operation_id='CreateImage',
        tags=['images'],
        responses={
            200: ImageUploadSerializer,
            400: CreateImageErrorResponseSerializer
        },
        examples=[OpenApiExample(
            status_codes=['400'],
            response_only=True,
            name="image-exists",
            value={'id': "3cc2e9e0-4172-44b1-8d65-bc3fffd076dc", 'hash': ["Image with this hash already exists"]},
            description="An error where an image with this hash already exists, including the existing image's UUID."
        )]
    )
)
class ImageCreate(CreateAPIView):
    """
    Create an image on the Arkindex image server, ready for upload to Amazon S3.

    The response includes an Amazon S3 URL that you can use to upload an image
    via HTTP PUT. Update the image's status to "checked" to confirm your image
    is successfully uploaded and visible to Arkindex.
    """

    serializer_class = ImageUploadSerializer
    permission_classes = (IsVerified, )
    scopes = (Scope.UploadS3Image, )

    def perform_create(self, serializer):
        image_hash = serializer.validated_data.get('hash')
        server = serializer.validated_data.get('server')
        path = serializer.validated_data.get('path')
        existing_image = server.images.filter(
            Q(hash=image_hash) | Q(path=path)
        ).first()

        if existing_image:
            error = {
                'id': str(existing_image.id),
            }
            if existing_image.hash == image_hash:
                error['hash'] = ['Image with this hash already exists']
            if existing_image.path == path:
                error['path'] = ['Image with this path already exists']
            raise ValidationError(error)

        width = serializer.validated_data.get('width')
        height = serializer.validated_data.get('height')
        if width and height and self.request.user.is_internal:
            # Async tasks may send the image's width and height;
            # we will assume that the image was checked by the task
            serializer.validated_data['status'] = S3FileStatus.Checked

        return serializer.save()


@extend_schema_view(post=extend_schema(operation_id='CreateIIIFURL', tags=['images']))
class IIIFURLCreate(CreateAPIView):
    """
    Create an image in Arkindex from an existing IIIF 2.x image by URL.
    The URL should be of the image's identifier, not of its Image Information request (`/info.json`).
    """
    serializer_class = IIIFImageSerializer
    permission_classes = (IsVerified, )
    scopes = (Scope.CreateIIIFImage, )


class IIIFInformationCreate(CreateAPIView):
    """
    Create an image in Arkindex from the result of an IIIF 2.x Image Information request.

    Image Information specification: https://iiif.io/api/image/2.1/#image-information
    """
    serializer_class = ImageInformationSerializer
    permission_classes = (IsVerified, )
    scopes = (Scope.CreateIIIFImage, )

    @extend_schema(
        operation_id='CreateIIIFInformation',
        # ImageInformationSerializer does not have any defined fields, so it is ignored in OpenAPI
        # This adds a schema that accepts any JSON object and causes a `requestBody` to be still present
        request=OpenApiTypes.OBJECT,
        responses={200: IIIFImageSerializer},
        tags=['images'],
    )
    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        serializer = IIIFImageSerializer(serializer.instance)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


@extend_schema(tags=['images'])
@extend_schema_view(
    get=extend_schema(description='Retrieve an image'),
    patch=extend_schema(description="Update an image's status"),
    put=extend_schema(description="Update an image's status")
)
class ImageRetrieve(RetrieveUpdateAPIView):
    """
    Get information about an image or trigger a check
    """

    permission_classes = (IsVerified, )
    serializer_class = ImageSerializer
    queryset = Image.objects.using('default').all()


@extend_schema_view(
    get=extend_schema(
        operation_id='ListImageElements',
        tags=['images'],
        parameters=[
            OpenApiParameter(
                'type',
                description='Filter elements by type',
                required=False,
            ),
            OpenApiParameter(
                'folder',
                description='Restrict to or exclude elements with folder types',
                type=bool,
                required=False,
            ),
        ]
    )
)
class ImageElements(ListAPIView):
    """
    List elements on an image
    """
    # For OpenAPI type discovery: an image's ID is in the path
    queryset = Image.objects.none()
    permission_classes = (IsVerified, )
    serializer_class = ElementSlimSerializer

    def get_queryset(self):
        filters = {
            'corpus__in': Corpus.objects.readable(self.request.user),
            'zone__image_id': self.kwargs['pk'],
        }
        if 'type' in self.request.query_params:
            filters['type__slug'] = self.request.query_params['type']

        only_folder = self.request.query_params.get('folder')
        if only_folder is not None:
            filters['type__folder'] = only_folder.lower() not in ('false', '0')

        return Element.objects.filter(**filters) \
            .prefetch_related('corpus', 'zone__image__server', 'type') \
            .order_by('type__slug', 'name', 'id')
