from enum import Enum
from urllib.parse import quote, unquote
from uuid import UUID

from django.contrib.gis.geos import Point
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers

from arkindex.images.models import POLYGON_MAX_POINTS
from arkindex.project.gis import ensure_linear_ring


class EnumField(serializers.ChoiceField):
    """
    Serializes an enum field into a JSON value
    """

    def __init__(self, enum, *args, **kwargs):
        assert issubclass(enum, Enum) or not enum
        self.enum = enum
        choices = [(item.value, item.name) for item in self.enum]
        super().__init__(choices, *args, **kwargs)

    def to_representation(self, obj):
        if not isinstance(obj, self.enum):
            obj = self.to_internal_value(obj)
        return obj.value

    def to_internal_value(self, data):
        assert self.enum is not None, "No enum set on EnumField"
        try:
            return self.enum(data)
        except ValueError:
            raise serializers.ValidationError('Value is not of type {}'.format(self.enum.__name__))


class PointField(serializers.ListField):
    child = serializers.IntegerField(min_value=0)

    def __init__(self, *args, **kwargs):
        kwargs.update(min_length=2, max_length=2)
        super().__init__(*args, **kwargs)

    def run_validation(self, data):
        try:
            return Point(*super().run_validation(data))
        except (AssertionError, TypeError, ValueError) as e:
            raise serializers.ValidationError(str(e))


class LinearRingField(serializers.ListField):
    child = PointField()

    def __init__(self, *args, **kwargs):
        # The maximum length comes from the maximum amount of points the (image_id, polygon) unique index can store:
        kwargs.update(min_length=3, max_length=POLYGON_MAX_POINTS)
        super().__init__(*args, **kwargs)

    def run_validation(self, data):
        value = super().run_validation(data)

        # Ensure the LineString is closed
        if value[0] != value[-1]:
            if len(value) == POLYGON_MAX_POINTS:
                raise serializers.ValidationError(
                    f'{POLYGON_MAX_POINTS} points are allowed only when the first and last points are equal.'
                )
            value.append(value[0])

        try:
            return ensure_linear_ring(value)
        except (AssertionError, ValueError) as e:
            raise serializers.ValidationError(str(e))


class IIIFPathField(serializers.CharField):
    """
    An IIIF image path field where special characters are escaped,
    including slashes.
    """

    def to_internal_value(self, data):
        value = super().to_internal_value(data)
        # If unquoting the value gives a different value, the value was already quoted
        if '/' in value or unquote(value) == value:
            value = quote(value, safe='')
        return value


@extend_schema_field({
    'oneOf': [
        {'type': 'string', 'format': 'uuid'},
        {'type': 'boolean'},
    ]
})
class BestClassField(serializers.UUIDField):

    def to_internal_value(self, data):
        try:
            return UUID(data)
        except (TypeError, ValueError, AttributeError):
            return str(data).lower() not in ('false', '0')
