from collections.abc import Iterable, Iterator, Sized

from django.db.models import CharField, Func
from django.urls import reverse

from arkindex.documents.models import Element


def build_absolute_url(element, request, name, id_argument='pk', **kwargs):
    """
    Build an absolute URL for a specified view using the element ID. Used by IIIF serializers.
    """
    kwargs[id_argument] = str(element.id)
    return request.build_absolute_uri(reverse(name, kwargs=kwargs))


class disconnect_signal():
    """
    Context manager to temporarily disconnect a signal
    """

    def __init__(self, signal, **kwargs):
        self.signal = signal
        self.kwargs = kwargs

    def __enter__(self):
        self.signal.disconnect(**self.kwargs)

    def __exit__(self, *args):
        self.signal.connect(**self.kwargs)


def build_tree(tree, *, corpus, type):
    """
    Build Element and ElementPath instances from a tree described by
    a dict associating elements with their direct ancestors.
    Should only be used for unit tests.

    {'child': ['mother', 'father'], 'grandchild': 'child'}

    Returns a dict associating element names with created Elements.
    """
    assert isinstance(tree, dict)

    def parse_value(val):
        if isinstance(val, str):
            return [val]
        return val

    # List all elements
    # Pick all the keys, then add elements in the tree's values
    # Removes the need to add 'element': [] for root elements
    names = set(tree.keys()).union(*[parse_value(val) for val in tree.values()])
    # Create elements
    elements = {
        name: Element.objects.create(
            corpus=corpus,
            name=name,
            type=type,
        )
        for name in names
    }
    # Create paths
    for child_name, value in tree.items():
        parent_names = parse_value(value)
        for parent_name in parent_names:
            elements[child_name].add_parent(elements[parent_name])
    return elements


def bounding_box(linestring):
    """
    Returns a 4-tuple (x, y, width, height) for the bounding box of a LineString or LinearRing
    """
    all_x, all_y = zip(*linestring.coords)
    x, y = min(all_x), min(all_y)
    width, height = max(all_x) - x, max(all_y) - y
    return int(x), int(y), int(width), int(height)


class BulkMap(Sized, Iterable):
    """
    Apply a function on the full list of items as soon as an iterator is requested.

    Useful to perform complex operations on a QuerySet with pagination
    without breaking through DRF's and Django's layers.
    """

    def __init__(self, func, iterable):
        assert callable(func), 'func must be callable'
        assert isinstance(iterable, Iterable), 'iterable must implement __iter__'
        assert isinstance(iterable, Sized), 'iterable must implement __len__'
        self.func = func
        self.iterable = iterable

    def __getitem__(self, key):
        assert isinstance(key, slice), 'BulkMap only supports slicing'
        return BulkMap(self.func, self.iterable[key])

    def __len__(self):
        # Proxy __len__ without calling the function
        if hasattr(self.iterable, 'count'):
            return self.iterable.count()
        else:
            return len(self.iterable)

    def __iter__(self):
        return iter(self.func(list(self.iterable)))


class CounterIterator(Sized, Iterator):
    """
    Iterate over any iterator while keeping count of how many items were iterated on.
    This allows accessing the total amount of items in an iterator once a full iteration is completed
    without storing any item in memory.
    """

    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.count = 0

    def __next__(self):
        # This can fail with StopIteration if the iteration ends.
        item = next(self.iterator)
        # It didn't fail, so we can increment the counter.
        self.count += 1
        return item

    def __len__(self):
        return self.count


class RTrimChr(Func):
    """
    Django's RTrim function only allows trimming spaces.
    This function allows trimming any character, not only spaces.
    """
    function = 'RTRIM'
    arity = 2
    output_field = CharField()
